# vim: ts=4:sw=4:expandtab

# This file is part of ReText
# Copyright: 2017-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt6.QtCore import QEvent, Qt
from PyQt6.QtGui import (
    QColor,
    QCursor,
    QDesktopServices,
    QFontInfo,
    QGuiApplication,
    QPalette,
    QTextDocument,
)
from PyQt6.QtWebEngineCore import (
    QWebEnginePage,
    QWebEngineSettings,
    QWebEngineUrlRequestInfo,
    QWebEngineUrlRequestInterceptor,
)
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtWidgets import QApplication, QLabel

from ReText import globalCache, globalSettings
from ReText.editor import getColor
from ReText.syncscroll import SyncScroll


class ReTextWebEngineUrlRequestInterceptor(QWebEngineUrlRequestInterceptor):
    def interceptRequest(self, info):
        if (info.resourceType() == QWebEngineUrlRequestInfo.ResourceType.ResourceTypeXhr
                and info.requestUrl().isLocalFile()):
            # For security reasons, disable XMLHttpRequests to local files
            info.block(True)


def str_rgba(color: QColor):
    """ Todo: More elegant use of QColor with alpha in stylesheet """
    return f"rgba({color.red()}, {color.green()}, {color.blue()}, {color.alpha()})"

class UrlPopup(QLabel):
    def __init__(self, window):
        super().__init__(window)
        self.window = window

        self.setStyleSheet('''
            border: 1px solid {:s};
            border-radius: 3px;
            background: {:s};
            '''.format(str_rgba(getColor('urlPopupBorder')),
                       str_rgba(getColor('urlPopupArea'))))
        self.fontHeight = self.fontMetrics().height()
        self.setVisible(False)

    def pop(self, url: str):
        """ Show link target on mouse hover

        QWebEnginePage emits signal 'linkHovered' which provides
        url: str -- target of link hovered (or empty on mouse-release)
        """
        if url:
            self.setText(url)
            windowBottom = self.window.rect().bottom()
            textWidth = self.fontMetrics().horizontalAdvance(url)
            self.setGeometry(-2, windowBottom-self.fontHeight-6,
                    textWidth+10, self.fontHeight+10)
            self.setVisible(True)
        else:
            self.setVisible(False)


class ReTextWebEnginePage(QWebEnginePage):
    def __init__(self, parent, tab):
        QWebEnginePage.__init__(self, parent)
        self.tab = tab
        self.interceptor = ReTextWebEngineUrlRequestInterceptor(self)
        self.setUrlRequestInterceptor(self.interceptor)
        self.urlPopup = UrlPopup(self.tab.p)
        self.linkHovered.connect(self.urlPopup.pop)

    def setScrollPosition(self, pos):
        self.runJavaScript(f"window.scrollTo({pos.x()}, {pos.y()});")

    def getPositionMap(self, callback):
        def resultCallback(result):
            if result:
                return callback({int(a): b for a, b in result.items()})

        script = """
        var elements = document.querySelectorAll('[data-posmap]');
        var result = {};
        var bodyTop = document.body.getBoundingClientRect().top;
        for (var i = 0; i < elements.length; ++i) {
            var element = elements[i];
            value = element.getAttribute('data-posmap');
            bottom = element.getBoundingClientRect().bottom - bodyTop;
            result[value] = bottom;
        }
        result;
        """
        self.runJavaScript(script, resultCallback)

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceId):
        print(f"level={level!r} message={message!r} lineNumber={lineNumber!r} sourceId={sourceId!r}")

    def acceptNavigationRequest(self, url, type, isMainFrame):
        if not isMainFrame:
            return True
        if url.scheme() == "data":
            return True
        if url.isLocalFile():
            localFile = url.toLocalFile()
            if localFile == self.tab.fileName:
                self.tab.startPendingConversion()
                return False
            if self.tab.openSourceFile(localFile):
                return False
        if globalSettings.handleWebLinks:
            return True
        QDesktopServices.openUrl(url)
        return False


class ReTextWebEnginePreview(QWebEngineView):

    def __init__(self, tab,
                 editorPositionToSourceLineFunc,
                 sourceLineToEditorPositionFunc):

        QWebEngineView.__init__(self, parent=tab)
        webPage = ReTextWebEnginePage(self, tab)

        handCursor = QCursor(Qt.CursorShape.PointingHandCursor)
        arrowCursor = QCursor(Qt.CursorShape.ArrowCursor)
        webPage.linkHovered.connect(lambda value: self.setCursor(handCursor if value else arrowCursor))

        self.setPage(webPage)

        self.syncscroll = SyncScroll(webPage,
                                     editorPositionToSourceLineFunc,
                                     sourceLineToEditorPositionFunc)
        self.editBox = tab.editBox

        settings = self.settings()
        settings.setDefaultTextEncoding('utf-8')
        settings.setAttribute(QWebEngineSettings.WebAttribute.LocalContentCanAccessRemoteUrls, True)
        if hasattr(QWebEngineSettings.WebAttribute, 'ForceDarkMode'):  # Qt >= 6.7
            # QGuiApplication.instance().styleHints().colorScheme() does not seem to
            # work properly on KDE Plasma, so let's re-use the same approach as our
            # editor uses.
            palette = QApplication.palette()
            windowColor = palette.color(QPalette.ColorRole.Window)
            if windowColor.lightness() <= 150:
                settings.setAttribute(QWebEngineSettings.WebAttribute.ForceDarkMode, True)

        # Events relevant to sync scrolling
        self.editBox.cursorPositionChanged.connect(self._handleCursorPositionChanged)
        self.editBox.verticalScrollBar().valueChanged.connect(self.syncscroll.handleEditorScrolled)
        self.editBox.resized.connect(self._handleEditorResized)

        # Scroll the preview when the mouse wheel is used to scroll
        # beyond the beginning/end of the editor
        self.editBox.scrollLimitReached.connect(self._handleWheelEvent)

    def setFont(self, font):
        settings = self.settings()
        settings.setFontFamily(QWebEngineSettings.FontFamily.StandardFont,
                               font.family())
        settings.setFontSize(QWebEngineSettings.FontSize.DefaultFontSize,
                             QFontInfo(font).pixelSize())

    def setHtml(self, html, baseUrl):
        # A hack to prevent WebEngine from stealing the focus
        self.setEnabled(False)
        super().setHtml(html, baseUrl)
        self.setEnabled(True)

    def _handleWheelEvent(self, event):
        # Only pass wheelEvents on to the preview if syncscroll is
        # controlling the position of the preview
        if self.syncscroll.isActive():
            QGuiApplication.sendEvent(self.focusProxy(), event)

    def event(self, event):
        # Work-around https://bugreports.qt.io/browse/QTBUG-43602
        if event.type() == QEvent.Type.ChildAdded:
            event.child().installEventFilter(self)
        elif event.type() == QEvent.Type.ChildRemoved:
            event.child().removeEventFilter(self)
        return super().event(event)

    def eventFilter(self, object, event):
        if event.type() == QEvent.Type.Wheel:
            if QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ControlModifier:
                self.wheelEvent(event)
                return True
        return False

    def findText(self, text, flags):
        options = QWebEnginePage.FindFlag(0)
        if flags & QTextDocument.FindFlag.FindBackward:
            options |= QWebEnginePage.FindFlag.FindBackward
        if flags & QTextDocument.FindFlag.FindCaseSensitively:
            options |= QWebEnginePage.FindFlag.FindCaseSensitively
        super().findText(text, options)
        return True

    def disconnectExternalSignals(self):
        self.editBox.cursorPositionChanged.disconnect(self._handleCursorPositionChanged)
        self.editBox.verticalScrollBar().valueChanged.disconnect(self.syncscroll.handleEditorScrolled)
        self.editBox.resized.disconnect(self._handleEditorResized)

        self.editBox.scrollLimitReached.disconnect(self._handleWheelEvent)

    def _handleCursorPositionChanged(self):
        editorCursorPosition = self.editBox.verticalScrollBar().value() + \
                       self.editBox.cursorRect().top()
        self.syncscroll.handleCursorPositionChanged(editorCursorPosition)

    def _handleEditorResized(self, rect):
        self.syncscroll.handleEditorResized(rect.height())

    def wheelEvent(self, event):
        if QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ControlModifier:
            newZoomFactor = globalCache.webEngineZoomFactor * (1.001 ** event.angleDelta().y())
            # Valid values are within the range from 0.25 to 5.0.
            globalCache.webEngineZoomFactor = max(min(newZoomFactor, 5.0), 0.25)
            self.setZoomFactor(globalCache.webEngineZoomFactor)
        return super().wheelEvent(event)

    def showEvent(self, event):
        self.setZoomFactor(globalCache.webEngineZoomFactor)
        return super().showEvent(event)



# This file is part of ReText
# Copyright: 2024-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFileSystemModel


class ReTextFileSystemModel(QFileSystemModel):

    def data(self, index, role):
        # Show file name in tooltip
        if role == Qt.ItemDataRole.ToolTipRole:
            role = QFileSystemModel.Roles.FileNameRole
        return super().data(index, role)



# vim: ts=4:sw=4:expandtab
#
# This file is part of ReText
# Copyright: 2014, 2017 Maurice van der Pot
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from markups import MarkdownMarkup, ReStructuredTextMarkup
from PyQt6.QtGui import QTextCursor

LARGER_THAN_ANYTHING = sys.maxsize

class Row:
    def __init__(self, block=None, text=None, separatorline=False, paddingchar=' '):
        self.block = block
        self.text = text
        self.separatorline = separatorline
        self.paddingchar = paddingchar

    def __repr__(self):
        return f"<Row '{self.text}' {self.separatorline} '{self.paddingchar}'>"

def _getTableLines(doc, pos, markupClass):
    startblock = doc.findBlock(pos)
    editedlineindex = 0
    offset = pos - startblock.position()

    rows = [ Row(block = startblock,
                 text = startblock.text()) ]

    block = startblock.previous()
    while any(c in block.text() for c in '+|'):
        rows.insert(0, Row(block = block,
                           text = block.text()))
        editedlineindex += 1
        block = block.previous()

    block = startblock.next()
    while any(c in block.text() for c in '+|'):
        rows.append(Row(block = block,
                        text = block.text()))
        block = block.next()

    if markupClass == MarkdownMarkup:
        for i, row in enumerate(rows):
            if i == 1:
                row.separatorline = True
                row.paddingchar = '-'
    elif markupClass == ReStructuredTextMarkup:
        for row in rows:
            if row.text.strip().startswith(('+-','+=')):
                row.separatorline = True
                row.paddingchar = row.text.strip()[1]
                row.text = row.text.replace('+', '|')
    return rows, editedlineindex, offset

# Modify the edited line to put the table borders after the edition in their original positions.
# It does not matter that this function changes the position of table borders before the edition,
# because table editing mode only ever changes the table to the right of the cursor position.
def _sortaUndoEdit(rows, editedlineindex, offset, editsize):
    aftertext = rows[editedlineindex].text
    if editsize < 0:
        beforetext = ' ' * -editsize + aftertext
    else:
        beforetext = aftertext[:offset] + aftertext[offset + editsize:]

    rows[editedlineindex].text = beforetext

# Given text and the position of the n-th edge, returns n - 1
def _getEdgeIndex(text, edge):
    return text[:edge].count('|')

def _determineRoomInCell(row, edge, edgeIndex, shrinking, startposition=0):

    if len(row.text) > edge and row.text[edge] == '|' and \
       (not edgeIndex or _getEdgeIndex(row.text, edge) == edgeIndex):
        clearance = 0
        cellwidth = 0
        afterContent = True
        for i in range(edge - 1, startposition - 1, -1):
            if row.text[i] == '|':
                break
            else:
                if row.text[i] == row.paddingchar and afterContent:
                    clearance += 1
                else:
                    afterContent = False
                cellwidth += 1

        if row.separatorline:
            if shrinking:
                # do not shrink separator cells below 3
                room = max(0, cellwidth - 3)
            else:
                # start expanding the cell if only the space for a right-align marker is left
                room = max(0, cellwidth - 1)
        else:
            room = clearance
    else:
        room = LARGER_THAN_ANYTHING

    return room

# Add an edit for a row to match the specified shift if it has an edge on the
# specified position
def _performShift(row, rowShift, edge, edgeIndex, shift):
    editlist = []

    # Any row that has an edge on the specified position and that doesn't
    # already have edits that shift it 'shift' positions, will get an
    # additional edit
    if len(row.text) > edge and row.text[edge] == '|' and rowShift != shift and \
       (not edgeIndex or _getEdgeIndex(row.text, edge) == edgeIndex):
        editsize = -(rowShift - shift)
        rowShift = shift

        # Insert one position further to the left on separator lines, because
        # there may be a space (for esthetical reasons) or an alignment marker
        # on the last position before the edge and that should stay next to the
        # edge.
        if row.separatorline:
            edge -= 1

        editlist.append((edge, editsize))

    return editlist, rowShift

# Finds the next edge position starting at offset in any row that is shifting.
# Rows that are not shifting when we are searching for an edge starting at
# offset, are rows that (upto offset) did not have any edges that aligned with
# shifting edges on other rows.
def _determineNextEdge(rows, rowShifts, offset):
    nextedge = None
    nextedgerow = None

    for row, rowShift in zip(rows, rowShifts):
        if rowShift != 0:
            edge = row.text.find('|', offset)
            if edge != -1 and (nextedge is None or edge < nextedge):
                nextedge = edge
                nextedgerow = row

    return nextedge, _getEdgeIndex(nextedgerow.text, nextedge) if nextedge else None

# Return a list of edits to be made in other lines to adapt the table lines to
# a single edit in the edited line.
def _determineEditLists(rows, editedlineindex, offset, editsize, alignWithAnyEdge):

    # rowShift represents how much the characters on a line will shift as a
    # result of the already collected edits to be made.
    rowShifts = [0 for _ in rows]
    rowShifts[editedlineindex] = editsize

    editLists = [[] for _ in rows]

    # Find the next edge position on the edited row
    currentedge, currentedgeindex = _determineNextEdge(rows, rowShifts, offset)
    firstEdge = True


    while currentedge:

        if alignWithAnyEdge:
            # Ignore what column the edge belongs to
            currentedgeindex = None

        if editsize < 0:
            # How much an edge shifts to the left depends on how much room
            # there is in the cells on any row that shares this edge.
            leastLeftShift = min(
                -rowShift + _determineRoomInCell(row, currentedge, currentedgeindex, True)
                for row, rowShift in zip(rows, rowShifts)
            )

            shift = max(editsize, -leastLeftShift)
        else:
            # When shifting right, determine how much only once based on how
            # much the edited cell needs to expand
            if firstEdge:
                room = _determineRoomInCell(
                    rows[editedlineindex],
                    currentedge,
                    currentedgeindex,
                    False,
                    offset,
                )
                shift = max(0, editsize - room)

        for i, row in enumerate(rows):
            editList, newRowShift = _performShift(
                row,
                rowShifts[i],
                currentedge,
                currentedgeindex,
                shift,
            )
            rowShifts[i] = newRowShift
            editLists[i].extend(editList)

        currentedge, currentedgeindex = _determineNextEdge(rows, rowShifts, currentedge + 1)
        firstEdge = False

    return editLists

def _performEdits(cursor, rows, editLists, linewithoffset, offset):
    cursor.joinPreviousEditBlock()
    for i, (row, editList) in enumerate(zip(rows, editLists)):

        for editpos, editsize in sorted(editList, reverse=True):

            if i == linewithoffset:
                editpos += offset

            cursor.setPosition(row.block.position() + editpos)
            if editsize > 0:
                cursor.insertText(editsize * row.paddingchar)
            else:
                for _ in range(-editsize):
                    cursor.deletePreviousChar()
    cursor.endEditBlock()

def adjustTableToChanges(doc, pos, editsize, markupClass):
    if markupClass in (MarkdownMarkup, ReStructuredTextMarkup):

        # This is needed because in ReSt cells can span multiple columns
        # and we can therefore not determine which edges in other rows
        # are supposed to be aligned with the edges in the edited row.
        alignWithAnyEdge = (markupClass == ReStructuredTextMarkup)

        rows, editedlineindex, offset = _getTableLines(doc, pos, markupClass)

        _sortaUndoEdit(rows, editedlineindex, offset, editsize)

        editLists = _determineEditLists(rows, editedlineindex, offset, editsize, alignWithAnyEdge)

        cursor = QTextCursor(doc)
        _performEdits(cursor, rows, editLists, editedlineindex, editsize)

def handleReturn(cursor, markupClass, newRow):
    if markupClass not in (MarkdownMarkup, ReStructuredTextMarkup):
        return False
    positionInBlock = cursor.positionInBlock()
    cursor.select(QTextCursor.SelectionType.BlockUnderCursor)
    oldLine = cursor.selectedText().lstrip('\u2029')
    if not ('| ' in oldLine or ' |' in oldLine):
        cursor.setPosition(cursor.block().position() + positionInBlock)
        return False
    indent = 0
    while oldLine[indent] in ' \t':
        indent += 1
    indentChars, oldLine = oldLine[:indent], oldLine[indent:]
    newLine = ''.join('|' if c in '+|' else ' ' for c in oldLine).rstrip()
    cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock)
    if newRow and markupClass == MarkdownMarkup:
        sepLine = ''.join(c if c in ' |' else '-' for c in oldLine)
        cursor.insertText('\n' + indentChars + sepLine)
    elif newRow:
        sepLine = ''.join('+' if c in '+|' else '-' for c in oldLine)
        cursor.insertText('\n' + indentChars + sepLine)
    cursor.insertText('\n' + indentChars + newLine)
    positionInBlock = min(positionInBlock, len(indentChars + newLine))
    cursor.setPosition(cursor.block().position() + positionInBlock)
    return True



# This file is part of ReText
# Copyright: 2012-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from enum import IntFlag, auto

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont, QSyntaxHighlighter, QTextCharFormat

from ReText.editor import getColor

reHtmlTags     = re.compile('<[^<>@]*>')
reHtmlSymbols  = re.compile(r'&#?\w+;')
reHtmlStrings  = re.compile('"[^"<]*"(?=[^<]*>)')
reHtmlComments = re.compile('<!--[^<>]*-->')
reAsterisks    = re.compile(r'(?<!\*)\*[^ \*][^\*]*\*')
reUnderline    = re.compile(r'(?<!_|\w)_[^_]+_(?!\w)')
reDblAsterisks = re.compile(r'(?<!\*)\*\*((?!\*\*).)*\*\*')
reDblUnderline = re.compile(r'(?<!_|\w)__[^_]+__(?!\w)')
reTrpAsterisks = re.compile(r'\*{3,3}[^\*]+\*{3,3}')
reTrpUnderline = re.compile('___[^_]+___')
reMkdHeaders   = re.compile('^#.+')
reMkdLinksImgs = re.compile(r'(?<=\[)[^\[\]]*(?=\])')
reMkdLinkRefs  = re.compile(r'(?<=\]\()[^\(\)]*(?=\))')
reBlockQuotes  = re.compile('^ *>.+')
reReSTDirects  = re.compile(r'\.\. [a-z]+::')
reReSTRoles    = re.compile('(:[a-z-]+:)(`.+?`)')
reReSTLinks    = re.compile('(`.+?<)(.+?)(>`__?)')
reReSTLinkRefs = re.compile(r'\.\. _`?(.*?)`?: (.*)')
reReSTFldLists = re.compile('^ *:(.*?):')
reTextileHdrs  = re.compile(r'^h[1-6][()<>=]*\.\s.+')
reTextileQuot  = re.compile(r'^bq\.\s.+')
reMkdCodeSpans = re.compile('`[^`]*`')
reMkdMathSpans = re.compile(r'\\[\(\[].*?\\[\)\]]')
reReSTCodeSpan = re.compile('``.+?``')
reWords        = re.compile('[^_\\W]+')
reSpacesOnEnd  = re.compile(r'\s+$')
reADocHeaders  = re.compile('^={1,5} .*')
reADocConsBold = re.compile(r'(?<!\w|\*)\*[^\*]+\*(?!\w|\*)')
reADocConsItal = re.compile(r'(?<!\w|_)_[^_]+_(?!\w|_)')
reADocUnCnItal = re.compile('__[^_]+__')
reADocConsCode = re.compile(r'(?<!\w|`)`[^`]+`(?!\w|`)')
reADocLinks    = re.compile(r'(https?://[^ ]+)(\[[^\]]+\])')
reADocDocAttr  = re.compile('^:.+?:')
reADocBlckAttr = re.compile(r'^\[.+?\]$')


class Formatter:
    def __init__(self, funcs=None):
        self._funcs = funcs or []

    def __or__(self, other):
        result = Formatter(self._funcs.copy())
        if isinstance(other, Formatter):
            result._funcs.extend(other._funcs)
        elif isinstance(other, QFont.Weight):
            result._funcs.append(lambda f: f.setFontWeight(other))
        return result

    def format(self, charFormat):
        for func in self._funcs:
            func(charFormat)

NF = Formatter()
ITAL = Formatter([lambda f: f.setFontItalic(True)])
UNDL = Formatter([lambda f: f.setFontUnderline(True)])

Weight = QFont.Weight

def FG(colorName):
    def func(f):
        return f.setForeground(getColor(colorName))
    return Formatter([func])

def QString_length(text):
    # In QString, surrogate pairs are represented using multiple QChars,
    # so the length of QString is not always equal to the number of graphemes
    # in it (which is the case with Python strings).
    return sum(2 if ord(char) > 65535 else 1 for char in text)


class Markup(IntFlag):
    Mkd = auto()
    ReST = auto()
    Textile = auto()
    ADoc = auto()
    HTML = auto()

    # Special value which means that no other markup is allowed inside this pattern
    CodeSpan = auto()


docTypesMapping = {
    'Markdown': Markup.Mkd,
    'reStructuredText': Markup.ReST,
    'Textile': Markup.Textile,
    'asciidoc': Markup.ADoc,
    'html': Markup.HTML,
}


class ReTextHighlighter(QSyntaxHighlighter):
    dictionaries = None
    docType = None

    patterns = (
        # regex,         color,                                markups
        (reMkdCodeSpans, FG('codeSpans'),                      Markup.Mkd | Markup.CodeSpan),
        (reMkdMathSpans, FG('codeSpans'),                      Markup.Mkd | Markup.CodeSpan),
        (reReSTCodeSpan, FG('codeSpans'),                      Markup.ReST | Markup.ADoc | Markup.CodeSpan),  # noqa: E501
        (reHtmlTags,     FG('htmlTags') | Weight.Bold,         Markup.Mkd | Markup.Textile | Markup.HTML),
        (reHtmlSymbols,  FG('htmlSymbols') | Weight.Bold,      Markup.Mkd | Markup.HTML),
        (reHtmlStrings,  FG('htmlStrings') | Weight.Bold,      Markup.Mkd | Markup.HTML),
        (reHtmlComments, FG('htmlComments'),                   Markup.Mkd | Markup.HTML),
        (reAsterisks,    ITAL,                                 Markup.Mkd | Markup.ReST),
        (reUnderline,    ITAL,                                 Markup.Mkd | Markup.Textile),
        (reDblAsterisks, NF | Weight.Bold,                     Markup.Mkd | Markup.ReST | Markup.Textile | Markup.ADoc),  # noqa: E501
        (reDblUnderline, NF | Weight.Bold,                     Markup.Mkd),
        (reTrpAsterisks, ITAL | Weight.Bold,                   Markup.Mkd),
        (reTrpUnderline, ITAL | Weight.Bold,                   Markup.Mkd),
        (reMkdHeaders,   FG('markdownHeaders') | Weight.Black, Markup.Mkd),
        (reMkdLinksImgs, FG('markdownLinks'),                  Markup.Mkd),
        (reMkdLinkRefs,  ITAL | UNDL,                          Markup.Mkd),
        (reBlockQuotes,  FG('blockquotes'),                    Markup.Mkd),
        (reReSTDirects,  FG('restDirectives') | Weight.Bold,   Markup.ReST),
        (reReSTRoles,    NF, FG('restRoles') | Weight.Bold, FG('htmlStrings'), Markup.ReST),
        (reTextileHdrs,  NF | Weight.Black,                    Markup.Textile),
        (reTextileQuot,  FG('blockquotes'),                    Markup.Textile),
        (reAsterisks,    NF | Weight.Bold,                     Markup.Textile),
        (reDblUnderline, ITAL,                                 Markup.Textile),
        (reReSTLinks,    NF, NF, ITAL | UNDL, NF,              Markup.ReST),
        (reReSTLinkRefs, NF, FG('markdownLinks'), ITAL | UNDL, Markup.ReST),
        (reReSTFldLists, NF, FG('restDirectives'),             Markup.ReST),
        (reADocHeaders,  FG('markdownHeaders') | Weight.Black, Markup.ADoc),
        (reADocConsBold, NF | Weight.Bold,                     Markup.ADoc),
        (reADocConsItal, ITAL,                                 Markup.ADoc),
        (reADocUnCnItal, ITAL,                                 Markup.ADoc),
        (reADocConsCode, FG('codeSpans'),                      Markup.ADoc | Markup.CodeSpan),
        (reADocLinks,    FG('markdownLinks'), ITAL | UNDL,     Markup.ADoc),
        (reADocDocAttr,  FG('htmlTags'),                       Markup.ADoc),
        (reADocBlckAttr, FG('htmlTags'),                       Markup.ADoc),
    )

    def highlightBlock(self, text):
        # Syntax highlighter
        codeSpans = set()
        if self.docType in docTypesMapping:
            markup = docTypesMapping[self.docType]
            for pattern, *formatters, markups in self.patterns:
                if not (markup & markups):
                    continue
                for match in pattern.finditer(text):
                    start, end = match.start(), match.end()
                    if markups & Markup.CodeSpan:
                        codeSpans.add((start, end))
                    elif any(start < codeEnd and end > codeStart
                             for codeStart, codeEnd in codeSpans):
                        # Ignore any syntax if its match intersects with code spans.
                        # See https://github.com/retext-project/retext/issues/529
                        continue
                    for i, formatter in enumerate(formatters):
                        charFormat = QTextCharFormat()
                        formatter.format(charFormat)
                        self.setFormat(QString_length(text[:match.start(i)]),
                                       QString_length(match.group(i)),
                                       charFormat)
        for match in reSpacesOnEnd.finditer(text):
            charFormat = QTextCharFormat()
            charFormat.setBackground(getColor('whitespaceOnEnd'))
            self.setFormat(QString_length(text[:match.start()]),
                           QString_length(match.group(0)),
                           charFormat)
        # Spell checker
        if self.dictionaries:
            charFormat = QTextCharFormat()
            charFormat.setUnderlineColor(Qt.GlobalColor.red)
            charFormat.setUnderlineStyle(QTextCharFormat.UnderlineStyle.SpellCheckUnderline)
            for match in reWords.finditer(text):
                finalFormat = QTextCharFormat()
                finalFormat.merge(charFormat)
                finalFormat.merge(self.format(match.start()))
                word = match.group(0)
                correct = any(dictionary.check(word) for dictionary in self.dictionaries)
                if not correct:
                    self.setFormat(QString_length(text[:match.start()]),
                                   QString_length(word),
                                   finalFormat)



# This file is part of ReText
# Copyright: 2015-2024 Dmitry Shachnev
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This is implementation of XSettings specification, described at
# <http://standards.freedesktop.org/xsettings-spec/xsettings-spec-0.5.html>

import ctypes
import ctypes.util
import struct


class _xcb_reply_t(ctypes.Structure):
    # this can be used instead of xcb_intern_atom_reply_t,
    # xcb_get_selection_owner_reply_t, etc
    _fields_ = [('response_type', ctypes.c_uint8),
                ('pad0',          ctypes.c_uint8),
                ('sequence',      ctypes.c_uint16),
                ('length',        ctypes.c_uint32),
                ('payload',       ctypes.c_uint32)]

class _xcb_cookie_t(ctypes.Structure):
    # this can be used instead of xcb_intern_atom_cookie_t,
    # xcb_get_selection_owner_cookie_t, etc
    _fields_ = [('sequence',      ctypes.c_uint)]

_xcb_error_messages = [
    None,
    'XCB error: socket, pipe and other stream error',
    'XCB connection closed: extension unsupported',
    'XCB connection closed: insufficient memory',
    'XCB connection closed: request length exceeded',
    'XCB connection closed: DISPLAY parse error',
    'XCB connection closed: invalid screen'
]

class XSettingsError(RuntimeError):
    pass

class XSettingsParseError(XSettingsError):
    pass

def get_raw_xsettings(display=0):
    # initialize the libraries
    xcb_library_name = ctypes.util.find_library('xcb')
    if xcb_library_name is None:
        raise XSettingsError('Xcb library not found')
    xcb = ctypes.CDLL(xcb_library_name)

    c_library_name = ctypes.util.find_library('c')
    if c_library_name is None:
        raise XSettingsError('C library not found')
    c = ctypes.CDLL(c_library_name)

    # set some args and return types
    c.free.argtypes = [ctypes.c_void_p]
    c.free.restype = None
    xcb.xcb_connect.argtypes = [ctypes.c_char_p, ctypes.POINTER(ctypes.c_int)]
    xcb.xcb_connect.restype = ctypes.c_void_p
    xcb.xcb_connection_has_error.argtypes = [ctypes.c_void_p]
    xcb.xcb_connection_has_error.restype = ctypes.c_int
    xcb.xcb_disconnect.argtypes = [ctypes.c_void_p]
    xcb.xcb_disconnect.restype = None
    xcb.xcb_intern_atom.argtypes = [ctypes.c_void_p, ctypes.c_uint8, ctypes.c_uint16, ctypes.c_char_p]
    xcb.xcb_intern_atom.restype = _xcb_cookie_t
    xcb.xcb_intern_atom_reply.argtypes = [ctypes.c_void_p, _xcb_cookie_t, ctypes.c_void_p]
    xcb.xcb_intern_atom_reply.restype = ctypes.POINTER(_xcb_reply_t)
    xcb.xcb_get_selection_owner.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
    xcb.xcb_get_selection_owner.restype = _xcb_cookie_t
    xcb.xcb_get_selection_owner_reply.argtypes = [ctypes.c_void_p, _xcb_cookie_t, ctypes.c_void_p]
    xcb.xcb_get_selection_owner_reply.restype = ctypes.POINTER(_xcb_reply_t)
    xcb.xcb_get_property.argtypes = [ctypes.c_void_p, ctypes.c_uint8, ctypes.c_uint32, ctypes.c_uint32,
                                     ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32]
    xcb.xcb_get_property.restype = _xcb_cookie_t
    xcb.xcb_get_property_reply.argtypes = [ctypes.c_void_p, _xcb_cookie_t, ctypes.c_void_p]
    xcb.xcb_get_property_reply.restype = ctypes.c_void_p
    xcb.xcb_get_property_value_length.argtypes = [ctypes.c_void_p]
    xcb.xcb_get_property_value_length.restype = ctypes.c_int
    xcb.xcb_get_property_value.argtypes = [ctypes.c_void_p]
    xcb.xcb_get_property_value.restype = ctypes.c_void_p

    # open the connection
    connection = xcb.xcb_connect(None, None)
    error = xcb.xcb_connection_has_error(connection)
    if error:
        raise XSettingsError(_xcb_error_messages[error])

    # get selection atom cookie
    buffer = f'_XSETTINGS_S{display}'.encode()
    cookie = xcb.xcb_intern_atom(connection, 0, len(buffer), buffer)

    # get selection atom reply
    reply = xcb.xcb_intern_atom_reply(connection, cookie, None)
    selection_atom = reply.contents.payload
    c.free(reply)

    # get selection owner cookie
    cookie = xcb.xcb_get_selection_owner(connection, selection_atom)

    # get selection owner reply
    reply = xcb.xcb_get_selection_owner_reply(connection, cookie, None)
    window = reply.contents.payload
    c.free(reply)

    # get settings atom cookie
    buffer = b'_XSETTINGS_SETTINGS'
    cookie = xcb.xcb_intern_atom(connection, 0, len(buffer), buffer)

    # get settings atom reply
    reply = xcb.xcb_intern_atom_reply(connection, cookie, None)
    settings_atom = reply.contents.payload
    c.free(reply)

    # get property cookie
    cookie = xcb.xcb_get_property(connection, 0, window, settings_atom, 0, 0, 0x2000)

    # get property reply
    reply = xcb.xcb_get_property_reply(connection, cookie, None)
    if reply is not None:
        length = xcb.xcb_get_property_value_length(reply)
        pointer = xcb.xcb_get_property_value(reply) if length else None
        result = ctypes.string_at(pointer, length)
        c.free(reply)

    # close the connection
    xcb.xcb_disconnect(connection)

    # handle possible errors
    if reply is None or not length:
        raise XSettingsError('XSettings not available')

    return result

def parse_xsettings(raw_xsettings):
    if len(raw_xsettings) < 12:
        raise XSettingsParseError('length < 12')

    if raw_xsettings[0] not in (0, 1):
        raise XSettingsParseError(f'wrong order byte: {raw_xsettings[0]}')
    byte_order = '<>'[raw_xsettings[0]]
    settings_count = struct.unpack(byte_order + 'I', raw_xsettings[8:12])[0]

    TypeInteger, TypeString, TypeColor = range(3)
    result = {}

    raw_xsettings = raw_xsettings[12:]
    offset = 0
    for i in range(settings_count):
        setting_type = raw_xsettings[offset]
        offset += 2
        name_length = struct.unpack(byte_order + 'H', raw_xsettings[offset:offset + 2])[0]
        offset += 2
        name = raw_xsettings[offset:offset + name_length]
        offset += name_length
        if offset & 3:
            offset += 4 - (offset & 3)
        offset += 4     # skip last-change-serial

        if setting_type == TypeInteger:
            value = struct.unpack(byte_order + 'I', raw_xsettings[offset:offset + 4])[0]
            offset += 4
        elif setting_type == TypeString:
            value_length = struct.unpack(byte_order + 'I', raw_xsettings[offset:offset + 4])[0]
            offset += 4
            value = raw_xsettings[offset:offset + value_length]
            offset += value_length
            if offset & 3:
                offset += 4 - (offset & 3)
        elif setting_type == TypeColor:
            value = struct.unpack(byte_order + 'HHHH', raw_xsettings[offset:offset + 8])
            offset += 8
        else:
            raise XSettingsParseError(f'Wrong setting type: {setting_type}')
        result[name] = value
    return result

def get_xsettings(display=0):
    raw_xsettings = get_raw_xsettings(display)
    return parse_xsettings(raw_xsettings)



#!/usr/bin/env python3
# vim: ts=4:sw=4:expandtab

# ReText
# Copyright 2011-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ctypes
import multiprocessing
import signal
import sys
from os.path import exists, isdir, join

import markups
from PyQt6.QtCore import (
    QCommandLineOption,
    QCommandLineParser,
    QFileInfo,
    QLibraryInfo,
    Qt,
    QTranslator,
)
from PyQt6.QtDBus import QDBusConnection, QDBusInterface
from PyQt6.QtNetwork import QNetworkProxyFactory
from PyQt6.QtWidgets import QApplication

from ReText import app_version, cache, globalSettings, packageDir, settings
from ReText.window import ReTextWindow


def canonicalize(option):
    if option == '-':
        return option
    return QFileInfo(option).canonicalFilePath()

def main():
    multiprocessing.set_start_method('spawn')

    if markups.__version_tuple__ < (2, ):
        sys.exit('Error: ReText needs PyMarkups 2.0 or newer to run.')

    try:
        # See https://github.com/retext-project/retext/issues/399
        # and https://launchpad.net/bugs/941826
        ctypes.CDLL('libGL.so.1', ctypes.RTLD_GLOBAL)
    except OSError:
        pass

    # Needed for Qt WebEngine on Windows
    QApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts)
    app = QApplication(sys.argv)
    app.setOrganizationName("ReText project")
    app.setApplicationName("ReText")
    app.setApplicationDisplayName("ReText")
    app.setApplicationVersion(app_version)
    app.setOrganizationDomain('mitya57.me')
    app.setDesktopFileName('me.mitya57.ReText')
    QNetworkProxyFactory.setUseSystemConfiguration(True)

    RtTranslator = QTranslator()
    RtTranslator.load('retext_' + globalSettings.uiLanguage,
                      join(packageDir, 'locale'))
    QtTranslator = QTranslator()
    translationsPath = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
    QtTranslator.load("qtbase_" + globalSettings.uiLanguage, translationsPath)
    app.installTranslator(RtTranslator)
    app.installTranslator(QtTranslator)

    parser = QCommandLineParser()
    parser.addHelpOption()
    parser.addVersionOption()
    previewOption = QCommandLineOption('preview',
        QApplication.translate('main', 'Open the files in preview mode'))
    newWindowOption = QCommandLineOption('new-window',
        QApplication.translate('main', 'Create a new window even if there is an existing one'))
    parser.addOption(previewOption)
    parser.addOption(newWindowOption)
    parser.addPositionalArgument('files',
        QApplication.translate('main', 'List of files to open'),
        '[files...]')

    parser.process(app)
    filesToOpen = parser.positionalArguments()

    print('Using configuration file:', settings.fileName())
    print('Using cache file:', cache.fileName())
    if globalSettings.appStyleSheet:
        with open(globalSettings.appStyleSheet) as sheetfile:
            app.setStyleSheet(sheetfile.read())
    window = ReTextWindow()

    # ReText can change directory when loading files, so we
    # need to have a list of canonical names before loading
    fileNames = list(map(canonicalize, filesToOpen))

    openInExistingWindow = (globalSettings.openFilesInExistingWindow
        and not parser.isSet(newWindowOption))
    connection = QDBusConnection.sessionBus()
    if connection.isConnected() and openInExistingWindow:
        connection.registerObject('/', window, QDBusConnection.RegisterOption.ExportAllSlots)
        serviceName = 'me.mitya57.ReText'
        if not connection.registerService(serviceName) and fileNames:
            print('Opening the file(s) in the existing window of ReText.')
            iface = QDBusInterface(serviceName, '/', '', connection)
            for fileName in fileNames:
                iface.call('openFileWrapper', fileName)
            qWidgetIface = QDBusInterface(serviceName, '/', 'org.qtproject.Qt.QWidget', connection)
            qWidgetIface.call('raise')
            sys.exit(0)

    window.show()
    readStdIn = False

    if globalSettings.openLastFilesOnStartup:
        window.restoreLastOpenedFiles()
    for fileName in fileNames:
        if isdir(fileName):
            window.fileSystemModel.setRootPath(fileName)
            window.treeView.setRootIndex(window.fileSystemModel.index(fileName))
            window.actionShowDirectoryTree.setChecked(True)
            window.treeView.setVisible(True)
        elif exists(fileName):
            window.openFileWrapper(fileName)
            if parser.isSet(previewOption):
                window.actionPreview.setChecked(True)
                window.preview(True)
        elif fileName == '-':
            readStdIn = True

    inputData = ''
    if readStdIn and sys.stdin is not None:
        if sys.stdin.isatty():
            print('Reading stdin, press ^D to end...')
        inputData = sys.stdin.read()
    if inputData or not window.tabWidget.count():
        window.createNew(inputData)
    signal.signal(signal.SIGINT, lambda sig, frame: window.close())
    sys.exit(app.exec())

if __name__ == '__main__':
    main()



'''
Position Map Extension for Python-Markdown
==========================================

This extension adds data-posmap attributes to the generated HTML elements that
can be used to relate HTML elements to the corresponding lines in the markdown
input file.

Note: the line number stored in the data-posmap attribute corresponds to the
      empty line *after* the markdown block that the HTML was generated from.

Copyright 2016 [Maurice van der Pot](griffon26@kfk4ever.com)
          2017-2024 Dmitry Shachnev

License: [BSD](http://www.opensource.org/licenses/bsd-license.php)

'''

import re
from xml.etree.ElementTree import SubElement

from markdown.blockprocessors import BlockProcessor
from markdown.extensions import Extension
from markdown.extensions.codehilite import CodeHilite
from markdown.preprocessors import Preprocessor
from markdown.util import HTML_PLACEHOLDER_RE

try:
    from pymdownx.highlight import Highlight
except ImportError:
    Highlight = None


POSMAP_MARKER_RE = re.compile(r'__posmapmarker__\d+\n\n')


class PosMapExtension(Extension):
    """ Position Map Extension for Python-Markdown. """

    def extendMarkdown(self, md):
        """ Insert the PosMapExtension blockprocessor before any other
            extensions to make sure our own markers, inserted by the
            preprocessor, are removed before any other extensions get confused
            by them.
        """
        md.preprocessors.register(PosMapMarkPreprocessor(md), 'posmap_mark', 50)
        md.preprocessors.register(PosMapCleanPreprocessor(md), 'posmap_clean', 5)
        md.parser.blockprocessors.register(PosMapBlockProcessor(md.parser), 'posmap', 150)

        # Monkey patch CodeHilite constructor to remove the posmap markers from
        # text before highlighting it
        orig_codehilite_init = CodeHilite.__init__

        def new_codehilite_init(self, src=None, *args, **kwargs):
            src = POSMAP_MARKER_RE.sub('', src)
            orig_codehilite_init(self, src=src, *args, **kwargs)
        CodeHilite.__init__ = new_codehilite_init

        # Same for PyMdown Extensions if it is available
        if Highlight is not None:
            orig_highlight_highlight = Highlight.highlight

            def new_highlight_highlight(self, src, *args, **kwargs):
                src = POSMAP_MARKER_RE.sub('', src)
                return orig_highlight_highlight(self, src, *args, **kwargs)
            Highlight.highlight = new_highlight_highlight


class PosMapMarkPreprocessor(Preprocessor):
    """ PosMapMarkPreprocessor - insert __posmapmarker__linenr entries at each empty line """

    def run(self, lines):
        new_text = []
        for i, line in enumerate(lines):
            new_text.append(line)
            if line == '':
                # Do not insert markers in the middle of an indented block.
                if 0 < i < len(lines) - 1:
                    # Find the closest lines before and after this one that have
                    # non-whitespace characters, and check if they are indented.
                    i_prev = i - 1
                    while i_prev > 0 and lines[i_prev].strip() == "":
                        i_prev -= 1
                    i_next = i + 1
                    while i_next < len(lines) - 1 and lines[i_next].strip() == "":
                        i_next += 1
                    if lines[i_prev].startswith(" ") and lines[i_next].startswith(" "):
                        continue

                new_text.append(f'__posmapmarker__{i}')
                new_text.append('')
        return new_text

class PosMapCleanPreprocessor(Preprocessor):
    """ PosMapCleanPreprocessor - remove __posmapmarker__linenr entries that
        accidentally ended up in the htmlStash. This could have happened
        because they were inside html tags or a fenced code block.
    """

    def run(self, lines):

        for i in range(self.md.htmlStash.html_counter):
            block = self.md.htmlStash.rawHtmlBlocks[i]
            block = re.sub(POSMAP_MARKER_RE, '', block)
            self.md.htmlStash.rawHtmlBlocks[i] = block

        return lines


class PosMapBlockProcessor(BlockProcessor):
    """ PosMapBlockProcessor - remove each marker and add a data-posmap
        attribute to the previous HTML element
    """

    def test(self, parent, block):
        return block.startswith('__posmapmarker__')

    def run(self, parent, blocks):
        block = blocks.pop(0)
        line_nr = block.split('__')[2]
        last_child = self.lastChild(parent)
        if last_child is not None:
            # Avoid setting the attribute on HTML placeholders, because it
            # would interfere with later replacement with literal HTML
            # fragments. In this case just add an empty <p> with the attribute.
            if last_child.text and re.match(HTML_PLACEHOLDER_RE, last_child.text):
                last_child = SubElement(parent, 'p')
            last_child.set('data-posmap', line_nr)

def makeExtension(*args, **kwargs):
    return PosMapExtension(*args, **kwargs)




# This file is part of ReText
# Copyright: 2015-2024 Dmitry Shachnev
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

def get_from_xsettings():
    from ReText.xsettings import XSettingsError, get_xsettings
    try:
        xsettings = get_xsettings()
    except XSettingsError:
        return
    if b'Net/IconThemeName' in xsettings:
        return xsettings[b'Net/IconThemeName'].decode()
    if b'Net/FallbackIconTheme' in xsettings:
        return xsettings[b'Net/FallbackIconTheme'].decode()

def get_from_gsettings():
    try:
        from gi.repository import Gio
    except ImportError:
        return
    schema = 'org.gnome.desktop.interface'
    if schema in Gio.Settings.list_schemas():
        settings = Gio.Settings.new(schema)
        return settings.get_string('icon-theme')

def get_from_gtk():
    try:
        from gi import require_version
        require_version('Gtk', '4.0')
        from gi.repository import Gtk
    except (ImportError, ValueError):
        return
    settings = Gtk.Settings.get_default()
    return settings.get_property('gtk-icon-theme-name')

def get_icon_theme():
    return (get_from_xsettings()
         or get_from_gsettings()
         or get_from_gtk())



# This file is part of ReText
# Copyright: 2013-2024 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from os.path import join

from markups.common import CONFIGURATION_DIR
from PyQt6.QtCore import QFile, QFileInfo, Qt, QUrl, pyqtSignal
from PyQt6.QtGui import QDesktopServices, QIcon
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from ReText import getBundledIcon, getSettingsFilePath, globalSettings
from ReText.icontheme import get_icon_theme

MKD_EXTS_FILE = join(CONFIGURATION_DIR, 'markdown-extensions.txt')

class FileDialogButton(QPushButton):
    def __init__(self, parent, fileName, label):
        icon = parent.parent.actIcon('document-open')
        super().__init__(icon, self.tr('Select'), parent)
        self.fileName = fileName
        self.label = label
        self.defaultText = self.tr('(none)')
        self.updateLabelText()
        self.clicked.connect(self.processClick)

    def processClick(self):
        raise NotImplementedError

    def updateLabelText(self):
        template = '<small>' + self.tr('Currently selected: %s') + '</small>'
        if self.fileName:
            self.label.setText(template % f'<a href="{self.fileName}">{self.fileName}</a>')
        else:
            self.label.setText(template % self.defaultText)

class FileSelectButton(FileDialogButton):
    def processClick(self):
        startDir = (QFileInfo(self.fileName).absolutePath()
                    if self.fileName else '')
        self.fileName = QFileDialog.getOpenFileName(
            self,
            self.tr('Select file to open'),
            startDir,
            self.tr('CSS files (*.css)') + ';;' + self.tr('All files (*)'),
        )[0]
        self.updateLabelText()

class DirectorySelectButton(FileDialogButton):
    def processClick(self):
        startDir = (QFileInfo(self.fileName).absolutePath()
                    if self.fileName else '')
        self.fileName = QFileDialog.getExistingDirectory(
            self, self.tr('Select directory to open'), startDir)
        self.updateLabelText()

class ClickableLabel(QLabel):
    clicked = pyqtSignal()

    def mousePressEvent(self, event):
        self.clicked.emit()
        super().mousePressEvent(event)


def setIconThemeFromSettings():
    QIcon.setThemeName(globalSettings.iconTheme)
    if QIcon.themeName() in ('hicolor', ''):
        if not QFile.exists(getBundledIcon('document-new')):
            QIcon.setThemeName(get_icon_theme())


class ConfigDialog(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.parent = parent
        self.initConfigOptions()
        self.layout = QVBoxLayout(self)
        path = getSettingsFilePath()
        pathLabel = QLabel(self.tr('Using configuration file at:') +
            f' <a href="{path}">{path}</a>', self)
        pathLabel.linkActivated.connect(self.openLink)
        self.layout.addWidget(pathLabel)
        self.tabWidget = QTabWidget(self)
        self.layout.addWidget(self.tabWidget)
        buttonBox = QDialogButtonBox(self)
        buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok |
            QDialogButtonBox.StandardButton.Apply | QDialogButtonBox.StandardButton.Cancel)
        buttonBox.accepted.connect(self.acceptSettings)
        buttonBox.button(QDialogButtonBox.StandardButton.Apply).clicked.connect(self.saveSettings)
        buttonBox.rejected.connect(self.close)
        self.initWidgets()
        self.configurators['rightMargin'].valueChanged.connect(self.handleRightMarginSet)
        rightMarginWrap = self.configurators['rightMarginWrap']
        if hasattr(rightMarginWrap, 'checkStateChanged'):  # Qt >= 6.7
            rightMarginWrap.checkStateChanged.connect(self.handleRightMarginWrapSet)
        else:
            rightMarginWrap.stateChanged.connect(
                lambda state: self.handleRightMarginWrapSet(Qt.CheckState(state))
            )
        self.layout.addWidget(buttonBox)

    def initConfigOptions(self):
        self.tabs = (
            (self.tr('Behavior'), (
                (self.tr('Automatically save documents'), 'autoSave'),
                (self.tr('Automatically open last documents on startup'), 'openLastFilesOnStartup'),
                (self.tr('Number of recent documents'), 'recentDocumentsCount'),
                (self.tr('Restore window geometry'), 'saveWindowGeometry'),
                (self.tr('Default preview state'), 'defaultPreviewState'),
                (self.tr('Open external links in ReText window'), 'handleWebLinks'),
                (self.tr('Markdown syntax extensions (comma-separated)'), 'markdownExtensions'),
                (self.tr('Enable synchronized scrolling for Markdown'), 'syncScroll'),
            #    (self.tr('Default Markdown file extension'), 'markdownDefaultFileExtension'),
            #    (self.tr('Default reStructuredText file extension'), 'restDefaultFileExtension'),
            )),
            (self.tr('Editor'), (
                (self.tr('Highlight current line'), 'highlightCurrentLine'),
                (self.tr('Show line numbers'), 'lineNumbersEnabled'),
                (self.tr('Line numbers are relative to current line'), 'relativeLineNumbers'),
                (self.tr('Tab key inserts spaces'), 'tabInsertsSpaces'),
                (self.tr('Tabulation width'), 'tabWidth'),
                (self.tr('Draw vertical line at column'), 'rightMargin'),
                (self.tr('Enable soft wrap'), 'rightMarginWrap'),
                (self.tr('Show document stats'), 'documentStatsEnabled'),
                (self.tr('Ordered list mode'), 'orderedListMode'),
            )),
            (self.tr('Interface'), (
                (self.tr('Hide toolbar'), 'hideToolBar'),
                (self.tr('Icon theme name'), 'iconTheme'),
                (self.tr('Stylesheet file'), 'styleSheet', True),
                (self.tr('Hide tabs bar when there is only one tab'), 'tabBarAutoHide'),
                (self.tr('Show full path in window title'), 'windowTitleFullPath'),
                (self.tr('Show directory tree'), 'showDirectoryTree'),
                (self.tr('Working directory'), 'directoryPath', True),
            ))
        )

    def initWidgets(self):
        self.configurators = {}
        for tabTitle, options in self.tabs:
            page = self.getPageWidget(options)
            self.tabWidget.addTab(page, tabTitle)

    def getPageWidget(self, options):
        page = QWidget(self)
        layout = QGridLayout(page)
        index = 0
        for option in options:
            displayname, name = option[:2]
            fileselector = option[2] if len(option) > 2 else False
            label = ClickableLabel(displayname + ':', self)
            if name == 'markdownExtensions':
                url = QUrl('https://github.com/retext-project/retext/wiki/Markdown-extensions')
                helpButton = QPushButton(self.tr('Help'), self)
                helpButton.clicked.connect(lambda: QDesktopServices.openUrl(url))
                layout.addWidget(label, index, 0)
                layout.addWidget(helpButton, index, 1)
                index += 1
                try:
                    extsFile = open(MKD_EXTS_FILE)
                    value = extsFile.read().rstrip().replace(extsFile.newlines, ', ')
                    extsFile.close()
                except Exception:
                    value = ''
                self.configurators[name] = QLineEdit(self)
                self.configurators[name].setText(value)
                layout.addWidget(self.configurators[name], index, 0, 1, 2)
                index += 1
                continue
            fileLabel = None
            if fileselector:
                fileLabel = QLabel(self)
                fileLabel.setIndent(15)
                fileLabel.linkActivated.connect(self.openLink)
            value = getattr(globalSettings, name)
            if name == 'defaultPreviewState':
                self.configurators[name] = QComboBox(self)
                self.configurators[name].addItem(self.tr('Editor'), 'editor')
                self.configurators[name].addItem(self.tr('Live preview'), 'live-preview')
                self.configurators[name].addItem(self.tr('Normal preview'), 'normal-preview')
                comboBoxIndex = self.configurators[name].findData(value)
                self.configurators[name].setCurrentIndex(comboBoxIndex)
            elif name == 'highlightCurrentLine':
                self.configurators[name] = QComboBox(self)
                self.configurators[name].addItem(self.tr('Disabled'), 'disabled')
                self.configurators[name].addItem(self.tr('Cursor Line'), 'cursor-line')
                self.configurators[name].addItem(self.tr('Wrapped Line'), 'wrapped-line')
                comboBoxIndex = self.configurators[name].findData(value)
                self.configurators[name].setCurrentIndex(comboBoxIndex)
            elif name == 'orderedListMode':
                self.configurators[name] = QComboBox(self)
                self.configurators[name].addItem(self.tr('Increment'), 'increment')
                self.configurators[name].addItem(self.tr('Repeat'), 'repeat')
                comboBoxIndex = self.configurators[name].findData(value)
                self.configurators[name].setCurrentIndex(comboBoxIndex)
            elif name == 'directoryPath':
                self.configurators[name] = DirectorySelectButton(self, value, fileLabel)
            elif isinstance(value, bool):
                self.configurators[name] = QCheckBox(self)
                self.configurators[name].setChecked(value)
                label.clicked.connect(self.configurators[name].nextCheckState)
            elif isinstance(value, int):
                self.configurators[name] = QSpinBox(self)
                if name == 'tabWidth':
                    self.configurators[name].setRange(1, 10)
                elif name == 'recentDocumentsCount':
                    self.configurators[name].setRange(5, 20)
                else:
                    self.configurators[name].setMaximum(200)
                self.configurators[name].setValue(value)
            elif isinstance(value, str) and fileselector:
                self.configurators[name] = FileSelectButton(self, value, fileLabel)
            elif isinstance(value, str):
                self.configurators[name] = QLineEdit(self)
                self.configurators[name].setText(value)
            layout.addWidget(label, index, 0)
            layout.addWidget(self.configurators[name], index, 1, Qt.AlignmentFlag.AlignRight)
            if fileLabel is not None:
                index += 1
                layout.addWidget(fileLabel, index, 0, 1, 2)
            index += 1
        return page

    def handleRightMarginSet(self, value):
        if value < 10:
            self.configurators['rightMarginWrap'].setChecked(False)

    def handleRightMarginWrapSet(self, state):
        if state == Qt.CheckState.Checked and self.configurators['rightMargin'].value() < 10:
            self.configurators['rightMargin'].setValue(80)

    def saveSettings(self):
        for name, configurator in self.configurators.items():
            if name == 'markdownExtensions':
                continue
            if isinstance(configurator, QCheckBox):
                value = configurator.isChecked()
            elif isinstance(configurator, QSpinBox):
                value = configurator.value()
            elif isinstance(configurator, QLineEdit):
                value = configurator.text()
            elif isinstance(configurator, QComboBox):
                value = configurator.currentData()
            elif isinstance(configurator, FileDialogButton):
                value = configurator.fileName
            setattr(globalSettings, name, value)
        self.applySettings()

    def applySettings(self):
        setIconThemeFromSettings()
        try:
            extsFile = open(MKD_EXTS_FILE, 'w')
            for ext in self.configurators['markdownExtensions'].text().split(','):
                if ext.strip():
                    extsFile.write(ext.strip() + '\n')
            extsFile.close()
        except Exception as e:
            print(e, file=sys.stderr)
        for tab in self.parent.iterateTabs():
            tab.editBox.updateFont()
            tab.editBox.setWrapModeAndWidth()
            tab.editBox.viewport().update()
        self.parent.updateStyleSheet()
        self.parent.tabWidget.setTabBarAutoHide(globalSettings.tabBarAutoHide)
        self.parent.toolBar.setVisible(not globalSettings.hideToolBar)
        self.parent.editBar.setVisible(not globalSettings.hideToolBar)
        path = globalSettings.directoryPath
        self.parent.fileSystemModel.setRootPath(path)
        self.parent.treeView.setRootIndex(self.parent.fileSystemModel.index(path))
        self.parent.actionShowDirectoryTree.setChecked(globalSettings.showDirectoryTree)
        self.parent.treeView.setVisible(globalSettings.showDirectoryTree)
        if globalSettings.autoSave:
            self.parent.autoSaveTimer.start(60000)
        else:
            self.parent.autoSaveTimer.stop()

    def acceptSettings(self):
        self.saveSettings()
        self.close()

    def openLink(self, link):
        QDesktopServices.openUrl(QUrl.fromLocalFile(link))
